class InvalidNameError(Exception):
    pass


class InvalidMoveError(Exception):
    pass


class InvalidBlockError(Exception):
    pass


class NodeUnavailable(Exception):
    pass


class OutOfRandomError(Exception):
    pass
